@@ -6,7 +6,14 @@ use std::path::PathBuf;
 /// A type to handle lines in a diff
 pub trait Diff {
     /// Set the file info
-    fn set_info(&mut self, old_name: &str, new_name: &str, op: FileOp, binary: bool);
+    fn set_info(
+        &mut self,
+        old_name: &str,
+        new_name: &str,
+        op: FileOp,
+        binary_sizes: Option<Vec<BinaryHunk>>,
+        file_mode: Option<FileMode>,
+    );
 
     /// Add a line in the diff
     ///
@@ -17,10 +24,27 @@ pub trait Diff {
     /// If a line is added ( ) then old_line is the line in the source file and new_line is the line in the destination file.
     fn add_line(&mut self, old_line: u32, new_line: u32, line: &[u8]);
 
+    /// A new hunk is created
+    fn new_hunk(&mut self);
+
     /// Close the diff: no more lines will be added
     fn close(&mut self);
 }
 
+/// An enum containing the size of binary hunks
+#[derive(Clone, Copy, Debug, PartialEq)]
+pub enum BinaryHunk {
+    Literal(usize),
+    Delta(usize),
+}
+
+/// File mode change
+#[derive(Clone, Copy, Debug, PartialEq)]
+pub struct FileMode {
+    pub old: u32,
+    pub new: u32,
+}
+
 /// A type to handle patch
 pub trait Patch<D: Diff> {
     /// Create a new diff where lines will be added
@@ -31,12 +55,12 @@ pub trait Patch<D: Diff> {
 }
 
 /// The different file operation
-#[derive(Debug, PartialEq)]
+#[derive(Clone, Copy, Debug, PartialEq)]
 pub enum FileOp {
-    /// The file is new
-    New,
-    /// The file is deleted
-    Deleted,
+    /// The file is new with its mode
+    New(u32),
+    /// The file is deleted with its mode
+    Deleted(u32),
     /// The file is renamed
     Renamed,
     /// The file is copied
@@ -45,6 +69,16 @@ pub enum FileOp {
     None,
 }
 
+impl FileOp {
+    pub fn is_new_or_deleted(self) -> bool {
+        match self {
+            FileOp::New(_) => true,
+            FileOp::Deleted(_) => true,
+            _ => false,
+        }
+    }
+}
+
 /// Type to read a patch
 pub struct PatchReader<'a> {
     buf: &'a [u8],
@@ -63,8 +97,7 @@ impl<'a> Debug for LineReader<'a> {
 
 impl<'a> LineReader<'a> {
     fn is_binary(&self) -> bool {
-        self.buf
-            == b"GIT binary patch"
+        self.buf == b"GIT binary patch"
     }
 
     fn is_rename_from(&self) -> bool {
@@ -93,9 +126,9 @@ impl<'a> LineReader<'a> {
 
     fn get_file_op(&self) -> FileOp {
         if self.is_new_file() {
-            FileOp::New
+            FileOp::New(self.parse_mode("new file mode "))
         } else if self.is_deleted_file() {
-            FileOp::Deleted
+            FileOp::Deleted(self.parse_mode("deleted file mode "))
         } else if self.is_rename_from() {
             FileOp::Renamed
         } else if self.is_copy_from() {
@@ -190,6 +223,21 @@ impl<'a> LineReader<'a> {
             std::str::from_utf8(new_path).unwrap(),
         )
     }
+
+    fn parse_mode(&self, start: &'static str) -> u32 {
+        // we know that line is beginning with "old/new mode "
+        // the following number is an octal number with 6 digits (so max is 8^6 - 1)
+        let buf = unsafe { self.buf.get_unchecked(start.len()..) };
+        let mut x = 0;
+        for c in buf.iter() {
+            if *c >= b'0' && *c <= b'7' {
+                x = x * 8 + u32::from(*c - b'0');
+            } else {
+                break;
+            }
+        }
+        x
+    }
 }
 
 impl<'a> PatchReader<'a> {
@@ -201,7 +249,7 @@ impl<'a> PatchReader<'a> {
                 reader.read_to_end(&mut data).unwrap();
                 PatchReader::by_buf(&data, patch);
             }
-            Err(_) => {
+            _ => {
                 panic!(format!("Failed to read the file: {:?}", path));
             }
         }
@@ -222,7 +270,7 @@ impl<'a> PatchReader<'a> {
 
     fn parse_diff<D: Diff, P: Patch<D>>(&mut self, diff_line: &mut LineReader, patch: &mut P) {
         trace!("Diff {:?}", diff_line);
-        let mut line = if let Some(line) = self.next(PatchReader::old_new_mode, false) {
+        let mut line = if let Some(line) = self.next(PatchReader::mv, false) {
             line
         } else {
             // Nothing more... so close it
@@ -231,13 +279,42 @@ impl<'a> PatchReader<'a> {
             trace!("Single diff line: new: {}", new);
 
             let diff = patch.new_diff();
-            diff.set_info(old, new, FileOp::None, false);
+            diff.set_info(old, new, FileOp::None, None, None);
             diff.close();
             return;
         };
+
+        let file_mode = if PatchReader::old_mode(&line) {
+            let old = line.parse_mode("old mode ");
+            let l = self.next(PatchReader::mv, false).unwrap();
+            let new = l.parse_mode("new mode ");
+            let file_mode = FileMode { old, new };
+            if let Some(l) = self.next(PatchReader::mv, false) {
+                line = l;
+                Some(file_mode)
+            } else {
+                // Nothing more... so close it
+                let (old, new) = diff_line.parse_files();
+
+                trace!("Single diff line (mode change): new: {}", new);
+
+                let diff = patch.new_diff();
+                diff.set_info(old, new, FileOp::None, None, Some(file_mode));
+                diff.close();
+                return;
+            }
+        } else {
+            None
+        };
+
         let op = line.get_file_op();
 
-        trace!("Diff (op = {:?}): {:?}, next_line: {:?}", op, diff_line, line);
+        trace!(
+            "Diff (op = {:?}): {:?}, next_line: {:?}",
+            op,
+            diff_line,
+            line
+        );
 
         if PatchReader::diff(&line) {
             let (old, new) = diff_line.parse_files();
@@ -245,7 +322,7 @@ impl<'a> PatchReader<'a> {
             trace!("Single diff line: old: {} -- new: {}", old, new);
 
             let diff = patch.new_diff();
-            diff.set_info(old, new, FileOp::None, false);
+            diff.set_info(old, new, FileOp::None, None, file_mode);
             diff.close();
             self.parse_diff(&mut line, patch);
             return;
@@ -270,7 +347,7 @@ impl<'a> PatchReader<'a> {
             trace!("Copy/Renamed from {} to {}", old, new);
 
             let diff = patch.new_diff();
-            diff.set_info(old, new, FileOp::Renamed, false);
+            diff.set_info(old, new, FileOp::Renamed, None, file_mode);
 
             if let Some(mut _line) = self.next(PatchReader::mv, false) {
                 if _line.is_triple_minus() {
@@ -287,9 +364,8 @@ impl<'a> PatchReader<'a> {
                     }
                 }
             }
-            return;
         } else {
-            if op == FileOp::New || op == FileOp::Deleted || line.is_index() {
+            if op.is_new_or_deleted() || line.is_index() {
                 trace!("New/Delete file: {:?}", line);
                 line = if let Some(line) = self.next(PatchReader::useful, false) {
                     line
@@ -300,7 +376,7 @@ impl<'a> PatchReader<'a> {
                     trace!("Single new/delete diff line: new: {}", new);
 
                     let diff = patch.new_diff();
-                    diff.set_info(old, new, op, false);
+                    diff.set_info(old, new, op, None, file_mode);
                     diff.close();
                     return;
                 };
@@ -313,9 +389,10 @@ impl<'a> PatchReader<'a> {
                     trace!("Binary file (op == {:?}): {}", op, new);
 
                     let diff = patch.new_diff();
-                    diff.set_info(old, new, op, true);
+                    let sizes = self.skip_binary();
+
+                    diff.set_info(old, new, op, Some(sizes), file_mode);
                     diff.close();
-                    self.skip_binary();
                     return;
                 } else if PatchReader::diff(&line) {
                     let (old, new) = diff_line.parse_files();
@@ -323,7 +400,7 @@ impl<'a> PatchReader<'a> {
                     trace!("Single new/delete diff line: new: {}", new);
 
                     let diff = patch.new_diff();
-                    diff.set_info(old, new, op, false);
+                    diff.set_info(old, new, op, None, file_mode);
                     diff.close();
                     self.parse_diff(&mut line, patch);
                     return;
@@ -346,7 +423,7 @@ impl<'a> PatchReader<'a> {
             trace!("Files: old: {} -- new: {}", old, new);
 
             let diff = patch.new_diff();
-            diff.set_info(old, new, op, false);
+            diff.set_info(old, new, op, None, file_mode);
             line = self.next(PatchReader::mv, false).unwrap();
             self.parse_hunks(&mut line, diff);
             diff.close();
@@ -365,6 +442,9 @@ impl<'a> PatchReader<'a> {
     fn parse_hunk<D: Diff>(&mut self, o: u32, n: u32, diff: &mut D) {
         let mut old_count = o;
         let mut new_count = n;
+
+        diff.new_hunk();
+
         while let Some(line) = self.next(PatchReader::hunk_change, true) {
             // we know that line is beginning with -, +, ... so no need to check bounds
             let first = unsafe { *line.buf.get_unchecked(0) };
@@ -435,25 +515,38 @@ impl<'a> PatchReader<'a> {
         self.pos = self.buf.len();
     }
 
-    fn skip_binary(&mut self) {
+    fn parse_usize(buf: &[u8]) -> usize {
+        let mut x = 0;
+        for c in buf.iter() {
+            if *c >= b'0' && *c <= b'9' {
+                x = x * 10 + usize::from(*c - b'0');
+            } else {
+                break;
+            }
+        }
+        x
+    }
+
+    fn skip_binary(&mut self) -> Vec<BinaryHunk> {
+        let mut sizes = Vec::new();
         loop {
             self.skip_until_empty_line();
             if let Some(buf) = self.buf.get(self.pos..) {
-                // 8 == len(literal )
-                if let Some(buf) = buf.get(..8) {
-                    if buf == b"literal " {
-                        continue;
-                    }
-                }
-                // 6 == len(delta )
-                if let Some(buf) = buf.get(..6) {
-                    if buf == b"delta " {
-                        continue;
-                    }
+                if buf.starts_with(b"literal ") {
+                    self.pos += 8;
+                    let buf = unsafe { self.buf.get_unchecked(self.pos..) };
+                    sizes.push(BinaryHunk::Literal(Self::parse_usize(buf)));
+                    continue;
+                } else if buf.starts_with(b"delta ") {
+                    self.pos += 6;
+                    let buf = unsafe { self.buf.get_unchecked(self.pos..) };
+                    sizes.push(BinaryHunk::Delta(Self::parse_usize(buf)));
+                    continue;
                 }
             }
             break;
         }
+        sizes
     }
 
     fn useful(line: &LineReader) -> bool {
@@ -472,8 +565,8 @@ impl<'a> PatchReader<'a> {
         line.buf.starts_with(b"@@ -")
     }
 
-    fn old_new_mode(line: &LineReader) -> bool {
-        !line.buf.starts_with(b"old ") && !line.buf.starts_with(b"new mode")
+    fn old_mode(line: &LineReader) -> bool {
+        line.buf.starts_with(b"old mode ")
     }
 
     fn hunk_change(line: &LineReader) -> bool {
@@ -499,7 +592,7 @@ mod tests {
             ("@@ -123,456 +789 @@", (123, 789)),
             ("@@ -123 +789 @@", (123, 789)),
         ];
-        for c in cases.into_iter() {
+        for c in cases.iter() {
             let buf = c.0.as_bytes();
             let line = LineReader { buf: &buf };
             assert!(line.parse_numbers() == c.1);
@@ -515,7 +608,7 @@ mod tests {
             ("   b/world/hello\t", "world/hello"),
             (" /dev/null\t", ""),
         ];
-        for c in cases.into_iter() {
+        for c in cases.iter() {
             let buf = c.0.as_bytes();
             let p = LineReader::get_filename(buf);
             assert!(p == c.1);
@@ -525,7 +618,7 @@ mod tests {
     #[test]
     fn test_line_starts_with() {
         let cases = [("+++ hello", "+++"), ("+++ hello", "++++")];
-        for c in cases.into_iter() {
+        for c in cases.iter() {
             let buf = c.0.as_bytes();
             let line = LineReader { buf: &buf };
             let pat = c.1.as_bytes();
@@ -556,7 +649,7 @@ mod tests {
             "abcdef",
             "ghijkl",
             "",
-            "delta 1",
+            "delta 2",
             "abcdef",
             "ghijkl",
             "",
@@ -568,8 +661,9 @@ mod tests {
             buf: s.as_bytes(),
             pos: 0,
         };
-        p.skip_binary();
-        assert!(p.pos == hpos);
+        let sizes = p.skip_binary();
+        assert_eq!(p.pos, hpos);
+        assert_eq!(sizes, vec![BinaryHunk::Literal(1), BinaryHunk::Delta(2)]);
     }
 
     #[test]
