@@ -27,6 +27,17 @@ pub struct LineChange {
     pub data: String,
 }
 
+#[derive(Default, Deserialize, PartialEq)]
+pub struct Hunk {
+    pub lines: Vec<LineChange>,
+}
+
+#[derive(Default, Deserialize, PartialEq)]
+pub struct FileModeChange {
+    pub old: u32,
+    pub new: u32,
+}
+
 #[derive(Deserialize)]
 pub struct DiffImpl {
     pub filename: String,
@@ -35,7 +46,8 @@ pub struct DiffImpl {
     pub binary: bool,
     pub copied_from: Option<String>,
     pub renamed_from: Option<String>,
-    pub lines: Vec<LineChange>,
+    pub hunks: Vec<Hunk>,
+    pub file_mode: Option<FileModeChange>,
 }
 
 #[derive(Deserialize)]
@@ -45,17 +57,39 @@ struct PatchImpl {
 
 impl Debug for LineChange {
     fn fmt(&self, f: &mut Formatter) -> Result {
-        write!(f, "line: {}, deleted: {}, data: {}", self.line, self.deleted, self.data)
+        write!(
+            f,
+            "line: {}, deleted: {}, data: {}",
+            self.line, self.deleted, self.data
+        )
     }
 }
 
 impl Debug for DiffImpl {
     fn fmt(&self, f: &mut Formatter) -> Result {
         writeln!(f, "filename: {}", self.filename)?;
-        writeln!(f, "new: {}, deleted: {}, binary: {}", self.new, self.deleted, self.binary)?;
-        writeln!(f, "copied_from: {:?}, renamed_from: {:?}", self.copied_from, self.renamed_from)?;
-        for line in self.lines.iter() {
-            writeln!(f, " - {:?}", *line)?;
+        writeln!(
+            f,
+            "new: {}, deleted: {}, binary: {}",
+            self.new, self.deleted, self.binary
+        )?;
+        writeln!(
+            f,
+            "copied_from: {:?}, renamed_from: {:?}",
+            self.copied_from, self.renamed_from
+        )?;
+        if let Some(file_mode) = &self.file_mode {
+            writeln!(
+                f,
+                "old mode: {:06o}, new mode: {:06o}",
+                file_mode.old, file_mode.new
+            )?;
+        }
+        for hunk in self.hunks.iter() {
+            writeln!(f, " @@")?;
+            for line in hunk.lines.iter() {
+                writeln!(f, " - {:?}", *line)?;
+            }
         }
         Ok(())
     }
@@ -79,7 +113,8 @@ impl Patch<DiffImpl> for PatchImpl {
             binary: false,
             copied_from: None,
             renamed_from: None,
-            lines: Vec::new(),
+            hunks: Vec::new(),
+            file_mode: None,
         });
         self.diffs.last_mut().unwrap()
     }
@@ -88,15 +123,24 @@ impl Patch<DiffImpl> for PatchImpl {
 }
 
 impl Diff for DiffImpl {
-    fn set_info(&mut self, old_name: &str, new_name: &str, op: FileOp, binary: bool) {
+    fn set_info(
+        &mut self,
+        old_name: &str,
+        new_name: &str,
+        op: FileOp,
+        binary_sizes: Option<Vec<BinaryHunk>>,
+        file_mode: Option<FileMode>,
+    ) {
         match op {
-            FileOp::New => {
+            FileOp::New(m) => {
                 self.new = true;
                 self.filename = new_name.to_string();
+                self.file_mode = Some(FileModeChange { old: 0, new: m });
             }
-            FileOp::Deleted => {
+            FileOp::Deleted(m) => {
                 self.deleted = true;
                 self.filename = old_name.to_string();
+                self.file_mode = Some(FileModeChange { old: m, new: 0 });
             }
             FileOp::Renamed => {
                 self.filename = new_name.to_string();
@@ -106,18 +150,27 @@ impl Diff for DiffImpl {
                 self.filename = new_name.to_string();
             }
         }
-        self.binary = binary;
+        self.binary = binary_sizes.is_some();
+
+        if self.file_mode.is_none() {
+            self.file_mode = file_mode.and_then(|c| {
+                Some(FileModeChange {
+                    old: c.old,
+                    new: c.new,
+                })
+            });
+        }
     }
 
     fn add_line(&mut self, old_line: u32, new_line: u32, line: &[u8]) {
         if old_line == 0 {
-            self.lines.push(LineChange {
+            self.hunks.last_mut().unwrap().lines.push(LineChange {
                 line: new_line,
                 deleted: false,
                 data: get_line(line),
             });
         } else if new_line == 0 {
-            self.lines.push(LineChange {
+            self.hunks.last_mut().unwrap().lines.push(LineChange {
                 line: old_line,
                 deleted: true,
                 data: get_line(line),
@@ -125,6 +178,10 @@ impl Diff for DiffImpl {
         }
     }
 
+    fn new_hunk(&mut self) {
+        self.hunks.push(Hunk::default());
+    }
+
     fn close(&mut self) {}
 }
 
@@ -144,11 +201,35 @@ fn get_line(buf: &[u8]) -> String {
 }
 
 fn compare(path: PathBuf, json: &PatchImpl, patch: &mut PatchImpl) {
-    patch.diffs.retain(|c| !c.binary);
+    if path == PathBuf::from("./tests/patches/6f5fc0d644dd.patch") {
+        let modes = patch.diffs[0].file_mode.as_ref().unwrap();
+        assert_eq!(modes.old, 0o100644);
+        assert_eq!(modes.new, 0o100755);
+    }
+
+    if path == PathBuf::from("./tests/patches/d8571fb523a0.patch") {
+        let modes = patch.diffs[1].file_mode.as_ref().unwrap();
+        assert_eq!(modes.new, 0o100644);
+        assert_eq!(modes.old, 0);
+    }
+
+    if path == PathBuf::from("./tests/patches/d7a700707ddb.patch") {
+        eprintln!("{:?}", patch);
+        for diff in patch.diffs.iter() {
+            if diff.filename == "layout/style/test/test_overscroll_behavior_pref.html" {
+                let modes = diff.file_mode.as_ref().unwrap();
+                assert_eq!(modes.new, 0);
+                assert_eq!(modes.old, 0o100644);
+            }
+        }
+    }
+
     assert!(
         json.diffs.len() == patch.diffs.len(),
-        "Not the same length in patch: {:?}",
-        path
+        "Not the same length in patch {:?}: {} ({} expected)",
+        path,
+        patch.diffs.len(),
+        json.diffs.len(),
     );
     for (cj, cp) in json.diffs.iter().zip(patch.diffs.iter()) {
         assert!(
@@ -166,18 +247,28 @@ fn compare(path: PathBuf, json: &PatchImpl, patch: &mut PatchImpl) {
             )
         );
         assert!(
-            cj.lines.len() == cp.lines.len(),
+            cj.hunks.len() == cp.hunks.len(),
             format!(
-                "Not the same length for changed lines: {} ({} expected)",
-                cp.lines.len(),
-                cj.lines.len()
+                "Not the same length for hunks: {} ({} expected)",
+                cp.hunks.len(),
+                cj.hunks.len()
             )
         );
-        for (lj, lp) in cj.lines.iter().zip(cp.lines.iter()) {
+        for (hj, hp) in cj.hunks.iter().zip(cp.hunks.iter()) {
             assert!(
-                lj == lp,
-                format!("Not the same line change: {:?} ({:?} expected", lp, lj)
+                hj.lines.len() == hp.lines.len(),
+                format!(
+                    "Not the same length for changed lines: {} ({} expected)",
+                    hp.lines.len(),
+                    hj.lines.len()
+                )
             );
+            for (lj, lp) in hj.lines.iter().zip(hp.lines.iter()) {
+                assert!(
+                    lj == lp,
+                    format!("Not the same line change: {:?} ({:?} expected", lp, lj)
+                );
+            }
         }
     }
 }
@@ -189,6 +280,9 @@ fn test_parse() {
         let entry = entry.unwrap();
         let path = entry.path();
         if !path.is_dir() && path.extension().unwrap() == "json" {
+            /*if path != PathBuf::from("./tests/output/6f5fc0d644dd.json") {
+                continue;
+            }*/
             let file = File::open(&path).unwrap();
             let reader = BufReader::new(file);
             let json_patch = serde_json::from_reader::<_, PatchImpl>(reader).unwrap();
